/* pgmtoppm [-m map] [file1 file2 ...] 
   Geen map - grijstinten */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>

#define COLORS 256

extern char *optarg;
extern int optind, opterr, optopt;

struct st_header
{
  int type;
  char comment[1024];
  int height,width,colors;
};

struct st_color
{
  unsigned char r,g,b;
};

struct st_color map[COLORS];

void usage()
{
  puts("Usage\npgmtoppm [-m map] [file file ...]");
  exit(1);
}

char *change_extension(char *old,char *new)
{
  int l;

  l=strlen(old);
  strcpy(new,old);
  if (l>4 && !strcasecmp(&old[l-4],".pgm"))
  {
    new[l-4]='\0';
  }
  strcat(new,".ppm");
  return(new);
}

int readmap(char *mapfile)
{
  int i;
  char line[1024],*p;
  FILE *fp;

  fp=fopen(mapfile,"r");
  if (fp)
  {
    i=0;
    while (fgets(line,sizeof(line),fp)!=NULL && i<COLORS) 
    {
      p=strtok(line," \t");
      if (p!=NULL)
      {
        map[i].r=atoi(p)%COLORS;
        p=strtok(NULL," \t");
        if (p!=NULL)
        {
          map[i].g=atoi(p)%COLORS;
          p=strtok(NULL," \t");
          if (p!=NULL)
          {
            map[i].b=atoi(p)%COLORS;
          }
        }
      }
      ++i;
    }
    fclose(fp);
  }
  else
  {
    fprintf(stderr,"Can't open %s\n[%d] %s\n",mapfile,errno,strerror(errno));
    exit(2);
  }
}

int fillmap()
{
  int i;

  for (i=0; i<COLORS; ++i)
  {
    map[i].r=map[i].g=map[i].b=i;
  }
}

struct st_header read_header(FILE *fp,char *filename)
{
  int lineno;
  char comment[1024],line[1024],*p;
  struct st_header header;

  for (lineno=0; lineno<4; ++lineno)
  {
    if (fgets(line,sizeof(line),fp)!=NULL) 
    {
      if (lineno==0)
      {
        if (!strncmp(line,"P2",2) || !strncmp(line,"P5",2))
        {
	  header.type=line[1]-'0';
        }
        else
        {
          fprintf(stderr,"%s: Not a pgm-file\n",filename);
          exit(3);
        }
      }
      if (lineno==1)
      {
        if (line[0]=='#')
	{
	  strcpy(header.comment,&line[2]);
	}
	else
	{
	  header.comment[0]='\0';
	}
      }
      if (lineno==2)
      {
        header.height=header.width=0;
        p=strtok(line," \t");
        if (p!=NULL)
        {
	  header.width=atoi(p);
          p=strtok(NULL," \t");
          if (p!=NULL)
          {
	    header.height=atoi(p);
          }
	  if (!header.height || !header.width)
          {
	    fprintf(stderr,"%s: Error getting metrics\n",filename);
	    exit(4);
          }
        }
      }
      if (lineno==3)
      {
        header.colors=atoi(line);
        if (!header.colors)
        {
	  fprintf(stderr,"%s: Error getting max colorvalue\n",filename);
	  exit(5);
        }
      }
    }
  }
  return(header);
}

int read_data2(FILE *fp,unsigned char *data, long size)
{
  long count;
  char line[1024],*p;

  count=0L;
  while (fgets(line,sizeof(line),fp)!=NULL) 
  {
    p=strtok(line," \t");
    while (count<size && p!=NULL)
    {
      data[count]=atoi(p);
      p=strtok(NULL," \t");
      ++count;
    }
  }
}

int read_data5(FILE *fp,unsigned char *data, long size)
{
  size_t count;
  
  count=fread(data,size,1,fp);
  return(count==(size_t) size);
}

unsigned char *read_data(FILE* fp,struct st_header header)
{
  long size;
  char *data;

  size=header.width*header.height;
  data=malloc(size);
  if (data==NULL)
  {
    fprintf(stderr,"read_data: Malloc failed.\n[%d] %s\n",errno,
	    strerror(errno));
    exit(1);
  }
  if (header.type==2)
  {
    read_data2(fp,data,size);
  }
  else
  {
    read_data5(fp,data,size);
  }
  return(data);
}

int write_output(FILE *fp,struct st_header header,unsigned char *data)
{
  int i,j,value;

  fputs("P6\n",fp);
  fprintf(fp,"# %s",header.comment);
  fprintf(fp,"%d %d\n",header.width,header.height);
  fprintf(fp,"%d\n",header.colors);
  for (i=0; i<header.height; ++i)
  {
    for (j=0; j<header.width; ++j)
    {
      value=data[i*header.width+j];
      fputc(map[value].r,fp);
      fputc(map[value].g,fp);
      fputc(map[value].b,fp);
    }
  }
}

/* pgmtoppm(FILE *fp_in,FILE *fp_out,char *filename) */
int pgmtoppm(FILE *fp_in,FILE *fp_out,char *filename)
{
  int colors=0,height,width,value;
  unsigned char *data;
  //*p;
  struct st_header header;

  header=read_header(fp_in,filename);
  /*printf("Type: %d\n",header.type);
  puts(header.comment);
  printf("%dx%d\n",header.width,header.height);
  printf("%d\n",header.colors);*/
  data=read_data(fp_in,header);
  write_output(fp_out,header,data);
  free(data);
  exit(45);
  /*
  while (fgets(line,sizeof(line),fp_in)!=NULL) 
  {
    if (lineno>3)
    {
      p=strtok(line," \t");
      while(p!=NULL)
      {
        value=atoi(p);
	fprintf(fp_out,"%d %d %d\n",map[value].r,map[value].g,map[value].b);
        p=strtok(NULL," \t");
      }
    }
    //puts(line);
    ++lineno;
  }
  */
  return(0);
}

main(int argc, char *argv[])
{
  int i,k;
  char mapfile[256],ppmfile[1024];
  FILE *fp_in,*fp_out;

  mapfile[0]='\0';
  do {
    k=getopt(argc,argv,"m:u");
    switch(k)
    {
      case 'm':
	strcpy(mapfile,optarg);
	break;
      case 'u':
	usage();
	break;
    }
  } while (k>-1);
  if (mapfile[0])
  {
    readmap(mapfile);
  }
  else
  {
    fillmap();
  }
  if (optind<argc)
  {
    for (i=optind; i<argc; ++i)
    {
      fp_in=fopen(argv[i],"r");
      if (fp_in!=NULL)
      {
        change_extension(argv[i],ppmfile);
	fp_out=fopen(ppmfile,"w");
	if (fp_out!=NULL)
	{
	  pgmtoppm(fp_in,fp_out,argv[i]);
	  fclose(fp_in);
	}
	else
	{
	  fprintf(stderr,"Can't open %s\n[%d] %s\n",ppmfile,errno,
		  strerror(errno));
	}
      }
      else
      {
	fprintf(stderr,"Can't open %s\n[%d] %s\n",argv[i],errno,
	        strerror(errno));
      }
    }
  }
  else
  {
    pgmtoppm(stdin,stdout,"stdin");
  }
}
